// CVWindow.cpp : implementation file
//

#include "stdafx.h"
#include "TPCut.h"
#include "CVWindow.h"
#include ".\cvwindow.h"
#include "MainFrm.h"
#include "CVCommon.h"
#include "resource.h"
#include <fstream>

// CVWindow

IMPLEMENT_DYNCREATE(CVWindow, CMDIChildWnd)

CVWindow::CVWindow()
{
}

CVWindow::CVWindow(char *pathname)
{
	img = cvLoadImage(pathname);
	firstFrm = NULL;
	width = img->width;
	height = img->height;
	cimg.CopyOf(img);
	double_img.Create(2*width,2*height,img->nChannels*8);
	half_img.Create(width/2,height/2,img->nChannels*8);

	ztype = NORMAL;
	zmfactor = 1;

	ROI.bottom = ROI.top = ROI.left = ROI.right = 0;
	ctype = NONE;
	disfun = NULL;
	rcoder = NULL;
	rcoder = new ROICoder(width, height, 1);

	is_showingMMap = false;
	actwesolver = NULL;

	mu =1.0; v = 0; lambda1= 1.0; lambda2=1.0; p=1; epsilon=1; dt = 10;

	stype = PDE;
	out_video = true;
	strcpy_s(vpath, "c:/out.avi");

	// Switch parameters
	once = true;
	first_time = true;
	first_chgparams = true;
	LBi = 0;   

	evolveTime = 0;
	
	startpnt.x = startpnt.y = 0;

	ax = NULL;
	mux = NULL;
	region_mask = NULL;

	actwesolver = new ActWESolver;

	cshape = RECTANGULAR;
	stype = GC;

	scale_factor = 1.0;

	show_img.CopyOf(img);

	ctype = SELECT;
}

CVWindow::~CVWindow()
{
	if(actwesolver){
		delete actwesolver;
	}	
	if(rcoder)
		delete rcoder;
	if(disfun)
		delete disfun;
	if(firstFrm)
		cvReleaseImage(&firstFrm);
	cvReleaseImage(&img);
}

BOOL CVWindow::Create(LPCTSTR szTitle, LONG style /*= 0*/ , 
				 const RECT& rect /* = rectDefault*/ , 
				 CMainFrame* pParent /*= NULL*/ )
{
	pParentFrm = pParent;
	LPCTSTR lpszHelloClass =
		AfxRegisterWndClass(CS_HREDRAW | CS_VREDRAW,
		LoadCursor(NULL, IDC_CROSS),
		(HBRUSH) (COLOR_WINDOW+1),
		LoadIcon(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDI_ICON_SNAKE)));
	return CMDIChildWnd::Create(lpszHelloClass, szTitle, style, rect, (CMDIFrameWnd*)pParent);
}

BEGIN_MESSAGE_MAP(CVWindow, CMDIChildWnd)
	ON_WM_PAINT()
	ON_WM_MOUSEMOVE()
	ON_WM_SIZE()
	ON_WM_LBUTTONDOWN()

	ON_COMMAND(ID_BUTTON_DRAW_CURVE, OnButtonDrawCurve)
	ON_COMMAND(ID_SHOW_DMAP, OnButtonShowMMap)
	//ON_COMMAND(ID_BUTTON_EVOLVE, OnButtonEvolveCurve)
	ON_COMMAND(ID_BUTTON_COMPUTE, OnButtonEvolveCurve)
	ON_COMMAND(ID_BUTTON_AUTOEVOLVE, OnButtonAutoEvolve)
	ON_COMMAND(ID_FILE_SAVE, OnSaveFile)
//	ON_COMMAND(IDC_BUTTON_CHANGE, OnEnChangeParams)
	ON_WM_TIMER()
	ON_WM_ERASEBKGND()	
	ON_WM_MOVING()
	ON_WM_SHOWWINDOW()
	ON_WM_CLOSE()
	
	ON_UPDATE_COMMAND_UI(ID_SHOW_DMAP, OnUpdateMButton)
	ON_WM_MOUSEWHEEL()
END_MESSAGE_MAP()


// CVWindow message handlers

void CVWindow::DrawCV(CDC* pdc)
{	
	CvvImage out_img;
//	if(ztype!=NORMAL){
		out_img.Create((int)(width*scale_factor), (int)(height*scale_factor), img->nChannels*8);
		cvResize(cimg.GetImage(),out_img.GetImage(), CV_INTER_NN);
		out_img.Show(pdc->GetSafeHdc(),WinROI.left,WinROI.top,(int)(scale_factor*width),(int)(scale_factor*height));
/*		if(ztype == TWICE){
			cvResize(cimg.GetImage(),double_img.GetImage(), CV_INTER_NN);
			double_img.Show(pdc->GetSafeHdc(),WinROI.left,WinROI.top,2*width,2*height);
		}else{
			cvResize(cimg.GetImage(),half_img.GetImage(), CV_INTER_NN);
			half_img.Show(pdc->GetSafeHdc(),WinROI.left,WinROI.top,width/2,height/2);
		}
*///	}else{
//		cimg.Show(pdc->GetSafeHdc(),WinROI.left,WinROI.top,cimg.GetImage()->width,cimg.GetImage()->height);
//	}
}

void CVWindow::OnPaint()
{
	CPaintDC dc(this); // device context for painting
	// TODO: Add your message handler code here
	// Do not call CMDIChildWnd::OnPaint() for painting messages	
	
	DrawCV(&dc);
	
}

void Saturate(CPoint &pnt, int w,int h)
{
	if(pnt.x < 0)  pnt.x = 0;
	if(pnt.y < 0)  pnt.y = 0;
	if(pnt.x >= w) pnt.x = w-1;
	if(pnt.y >= h) pnt.y = h-1;
}
void CVWindow::OnMouseMove(UINT nFlags, CPoint point)
{
	// TODO: Add your message handler code here and/or call default
	CPoint impnt;
// 	impnt.x = (int)((point.x-WinROI.left)/zmfactor);
// 	impnt.y = (int)((point.y-WinROI.top)/zmfactor);
	impnt.x = (int)((point.x-WinROI.left)/scale_factor);
	impnt.y = (int)((point.y-WinROI.top)/scale_factor);
	if(WinROI.PtInRect(point)){
		pParentFrm->SetXCoord(impnt.x);
		pParentFrm->SetYCoord(impnt.y);
	}else{
		pParentFrm->SetXCoord(-1);
		pParentFrm->SetYCoord(-1);
	}

	if(ctype!=NONE){
		if(rcoder->curstate == ROICoder::state::PROCEED ){
			Saturate(impnt, width, height);
			if(cshape == FREEFORM){
				rcoder->AddContourNode(cvPoint(impnt.x, impnt.y));
				CvSeq *seq = rcoder->getContour(rcoder->getCNum());
				int n = seq->total;
				for(int i=0; i<n; i++){
					CvPoint *pnt = (CvPoint*)cvGetSeqElem(seq, i);
					SetImageIntValue(cimg.GetImage(),pnt->x,pnt->y,0,0);
					SetImageIntValue(cimg.GetImage(),pnt->x,pnt->y,255,1);
					SetImageIntValue(cimg.GetImage(),pnt->x,pnt->y,255,2);
				}
			}else{				
				cimg.CopyOf(show_img.GetImage());				
				cvRectangle(cimg.GetImage(),startpnt,cvPoint(impnt.x,impnt.y),cvScalar(0,255,255));
			}
			this->Invalidate(false);
		}
	}

	CMDIChildWnd::OnMouseMove(nFlags, point);
}

void CVWindow::OnSize(UINT nType, int cx, int cy)
{
	CMDIChildWnd::OnSize(nType, cx, cy);
	winx = cx;
	winy = cy;
	// TODO: Add your message handler code here
	if(cx > width){
		ROI.left = (cx - width) / 2;		
		ROI.right = ROI.left + width;
	}else{
		ROI.left = 0;
		ROI.right = width;
	}
	if(cy > height){
		ROI.top = (cy - height) / 2;
		ROI.bottom = ROI.top + height;
	}else{
		ROI.top = 0;
		ROI.bottom = height;
	}	
	WinROI.left = (ROI.left+ROI.right)/2 -  (LONG)(width*scale_factor)/2;
	WinROI.right = WinROI.left +  (LONG)(scale_factor*width);
	WinROI.top = (ROI.top+ROI.bottom)/2 -  (LONG)(scale_factor*height)/2;
	WinROI.bottom = WinROI.top +  (LONG)(scale_factor*height);
}

void CVWindow::OnLButtonDown(UINT nFlags, CPoint point)
{
	// TODO: Add your message handler code here and/or call default
	CPoint impnt;
// 	impnt.x = (int)((point.x-WinROI.left)/zmfactor);
// 	impnt.y = (int)((point.y-WinROI.top)/zmfactor);
	impnt.x = (int)((point.x-WinROI.left)/scale_factor);
	impnt.y = (int)((point.y-WinROI.top)/scale_factor);

	
 	if(ctype == NONE)
 		return;
 	else{
		if(rcoder->curstate == ROICoder::state::BEGIN){
			rcoder->BeginContour(cvPoint(impnt.x, impnt.y));
			startpnt.x = impnt.x;
			startpnt.y = impnt.y;			
		}
		else if(rcoder->curstate == ROICoder::state::PROCEED){
			if(cshape == RECTANGULAR){
				rcoder->AddContourNode(cvPoint(startpnt.x, impnt.y));
				rcoder->AddContourNode(cvPoint(impnt.x, impnt.y));
				rcoder->AddContourNode(cvPoint(impnt.x, startpnt.y));	

				key_points.x[0] = startpnt.x; key_points.y[0] = startpnt.y;
				key_points.x[1] = startpnt.x; key_points.y[1] = impnt.y;
				key_points.x[2] = impnt.x; key_points.y[2] = impnt.y;
				key_points.x[3] = impnt.x; key_points.y[3] = startpnt.y;	

				// Debug
// 				std::ofstream out("c:/pos.txt");
							
// 				out << key_points.x[0] << " " << key_points.y[0] << endl;
// 				out << key_points.x[1] << " " << key_points.y[1] << endl;
// 				out << key_points.x[2] << " " << key_points.y[2] << endl;
// 				out << key_points.x[3] << " " << key_points.y[3] << endl;

				contours.push_back(key_points);
			}
			rcoder->EndContour();
			CvSeq *seq = rcoder->getContour(rcoder->getCNum() - 1);			
			int n = seq->total;
			for(; LBi<n; LBi++){
				CvPoint *pnt = (CvPoint*)cvGetSeqElem(seq, LBi);
				SetImageIntValue(cimg.GetImage(),pnt->x,pnt->y,0,0);
				SetImageIntValue(cimg.GetImage(),pnt->x,pnt->y,255,1);
				SetImageIntValue(cimg.GetImage(),pnt->x,pnt->y,255,2);
			}	

			cvCopyImage(cimg.GetImage(), show_img.GetImage());
		
			this->Invalidate(false);

			firstFrm = cvCloneImage(cimg.GetImage());			
		}else{
			//delete rcoder;
			//rcoder = new ROICoder(width, height, 1);
		}

	}
	CMDIChildWnd::OnLButtonDown(nFlags, point);
}

void CVWindow::OnButtonDrawCurve()
{
	if(ctype == NONE){
		ctype = SELECT;		
	}
}

void CVWindow::OnButtonShowDMap()
{	
	if(ctype != NONE){
		if (rcoder->curstate == rcoder->BEGIN && rcoder->getCNum() > 0){
			if(once){
				disfun = new DistanceFun(rcoder);
				if(stype == PDE)
					disfun->CalcuDisFun(0);
 				else
					disfun->CalcuDisFun(1);
				once = false;
			}

// 			for(int i=0; i<cimg.Width(); i++){
// 				for(int j=0; j<cimg.Height(); j++){
// 					int val = abs((int)(disfun->GetDisFun(i,j)/disfun->maxvalue));
// 					SetImageIntValue(cimg.GetImage(),i,j,val,0);
// 					SetImageIntValue(cimg.GetImage(),i,j,val,1);
// 					SetImageIntValue(cimg.GetImage(),i,j,val,2);					
// 
// 					IplImage *rmask = disfun->rcoder->GetRegionMask();
// 					if(GetSingleIntValue(rmask, i, j) == 255){
// 						SetImageIntValue(cimg.GetImage(),i,j,255,0);
// 						SetImageIntValue(cimg.GetImage(),i,j,255,1);
// 						SetImageIntValue(cimg.GetImage(),i,j,255,2);
// 					}
// 				}
// 			}

//			Invalidate(false);				
		}			
	}		
}

void CVWindow::OnButtonShowMMap()
{
	if(disfun == NULL)
		return;
	if(!is_showingMMap) {
		for(int i=0; i<cimg.Width(); i++){
			for(int j=0; j<cimg.Height(); j++){
				if(actwesolver->what_segment(i, j) < 0){
					SetImageIntValue(cimg.GetImage(),i,j,255,0);
					SetImageIntValue(cimg.GetImage(),i,j,255,1);
					SetImageIntValue(cimg.GetImage(),i,j,255,2);
				}else{
					SetImageIntValue(cimg.GetImage(),i,j,0,0);
					SetImageIntValue(cimg.GetImage(),i,j,0,1);
					SetImageIntValue(cimg.GetImage(),i,j,0,2);
				}
			}
		}
		is_showingMMap = 1;
	}else{
		actwesolver->Extract_Contour(actwesolver->GetResult(), region_mask, 0);
		cimg.CopyOf(img);
		for(int i=0; i<width; i++){
			for(int j=0; j<height; j++){
				double val = cvmGet(region_mask, i,j);
				if(val > 0.5){					
					SetImageIntValue(cimg.GetImage(), i,j,0,0);
					SetImageIntValue(cimg.GetImage(), i,j,255,1);
					SetImageIntValue(cimg.GetImage(), i,j,255,2);
				}
			}
		}
		is_showingMMap = 0;
	}

	Invalidate(false);		
}

void Gradient(CvMat *I, CvMat *gradIx, CvMat *gradIy)
{
	int width = I->height;
	int height = I->width;
	double val;
	for(int i=0; i<width; i++)
		for(int j=0; j<height; j++){
			// For gradient x
			if(i>0 && i<(width-1)){
				val = (cvmGet(I, i+1, j) - cvmGet(I, i-1, j))/2.0;				
			}else if(i == 0){
				val = cvmGet(I, i+1, j) - cvmGet(I, i, j);
			}else{ // i==width -1
				val = cvmGet(I, i, j) - cvmGet(I, i-1, j);
			}
			cvmSet(gradIx, i, j, val);

			// For gradient y
			if(j>0 && j<(height-1)){
				val = (cvmGet(I, i, j+1) - cvmGet(I, i, j-1))/2.0;				
			}else if(j == 0){
				val = cvmGet(I, i, j+1) - cvmGet(I, i, j);
			}else{ // j==height -1
				val = cvmGet(I, i, j) - cvmGet(I, i, j-1);
			}
			cvmSet(gradIy, i, j, val);
		}
}

void CVWindow::OnButtonEvolveCurve()
{
	OnButtonShowDMap();

	if(disfun == NULL || ctype == NONE){		
		//AfxMessageBox("Calculate distance function first.");
		return;
	}

	
	if(first_time){
		region_mask = cvCreateMat( width, height, CV_32FC1);
		CvMat *mat = cvCreateMat(width, height, CV_32FC1);
		Image2mat(img, mat, 0);
//		actwesolver = new ActWESolver;
   
		actwesolver->loadParam(disfun->GetDisFun(), mat, 1.00, v, 10, 10, p, epsilon, 10, 10, out_video,vpath);

		while(!contours.empty()) {
			key_points = contours.front();
			actwesolver->loadContour(key_points.y, key_points.x, 4);
			contours.pop_front();
		}
		actwesolver->output2movie(firstFrm);

		first_time = false;		

		cvReleaseMat(&mat);
	}	

	cimg.CopyOf(img);
	{

		{
			if(stype == PDE)
				actwesolver->Iterate();
			else
				actwesolver->IterateGC();

			actwesolver->Extract_Contour(actwesolver->GetResult(), region_mask, 0);
//			cvCopy(actwesolver->GetResult(),region_mask);
			for(int i=0; i<width; i++){
				for(int j=0; j<height; j++){
					double val = cvmGet(region_mask, i,j);
					if(val > 0.5){
						SetImageIntValue(cimg.GetImage(), i,j,0,0);
						SetImageIntValue(cimg.GetImage(), i,j,255,1);
						SetImageIntValue(cimg.GetImage(), i,j,255,2);
					}/*
					else{
						SetImageIntValue(cimg.GetImage(), i,j,0,0);
						SetImageIntValue(cimg.GetImage(), i,j,0,1);
						SetImageIntValue(cimg.GetImage(), i,j,0,2);
					}*/
					
				}
			}
			Invalidate(false);		

			//actwesolver->output2movie(cimg.GetImage());

			evolveTime ++ ;
			pParentFrm->SetFrmNum(actwesolver->iterate_num);
			pParentFrm->SetFrmTime(actwesolver->elapsetime_ms);
			pParentFrm->SetTotalTime(actwesolver->totol_elapsetime_ms);
		}
	}

	ctype = NONE;
}

void CVWindow::OnButtonAutoEvolve()
{
}
void CVWindow::OnTimer(UINT nIDEvent)
{
	// TODO: Add your message handler code here and/or call default
	CMDIChildWnd::OnTimer(nIDEvent);
}

BOOL CVWindow::OnEraseBkgnd(CDC* pDC)
{
	// TODO: Add your message handler code here and/or call default

	static CBrush brush(RGB (200 , 180 , 180) );
	// Select the brush into the device context .
	CBrush* pOldBrush = pDC->SelectObject (&brush);
	CRect rcClip;
	pDC->GetClipBox (&rcClip);//Paint the area.
	pDC->PatBlt (rcClip.left , rcClip.top ,rcClip.Width() , rcClip.Height() ,PATCOPY );
	pDC->SelectObject (pOldBrush );
	return TRUE;

	return CMDIChildWnd::OnEraseBkgnd(pDC);
}

class CFileWriteDlg : public CDialog
{
public:
	CFileWriteDlg();

	// Dialog Data
	enum { IDD = IDD_FILE_WRITE };
protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	// Implementation
protected:
	DECLARE_MESSAGE_MAP()
};

CFileWriteDlg::CFileWriteDlg() : CDialog(CFileWriteDlg::IDD)
{
}

void CFileWriteDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
}

BEGIN_MESSAGE_MAP(CFileWriteDlg, CDialog)
END_MESSAGE_MAP()

void CVWindow::OnSaveFile()
{
	char BASED_CODE szFilter[] ="BMP Files (*.bmp)|*.bmp|JPEG Files (*.jpg)|*.jpg|";
	CFileDialog dlg( FALSE, "bmp", NULL, OFN_HIDEREADONLY,szFilter);
	if(dlg.DoModal()!=IDOK) 
		return;
	CFile file;	
	if(file.Open(dlg.GetPathName(), CFile::modeRead)!=0){
		CFileWriteDlg dialog;		
		if(dialog.DoModal()!=IDOK)
			return;
	}else{
		cvSaveImage(dlg.GetPathName().GetBuffer(), cimg.GetImage());
	}
}

void CVWindow::OnClose()
{
	// TODO: Add your message handler code here and/or call default
	CMDIChildWnd::OnClose();
}

void CVWindow::OnUpdateMButton(CCmdUI *pCmdUI)
{
	if(is_showingMMap)
		pCmdUI->SetCheck(1);
	else
		pCmdUI->SetCheck(0);
}


BOOL CVWindow::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt)
{
	// TODO: Add your message handler code here and/or call default
	if(zDelta > 0) {
		scale_factor *= 1.2;
		if(scale_factor > 10)
			scale_factor = 10;
	}else{
		scale_factor /= 1.2;
		if(scale_factor < 0.1)
			scale_factor = 0.1;
	}

	WinROI.left = (ROI.left+ROI.right)/2 -  (LONG)(width*scale_factor)/2;
	WinROI.right = WinROI.left + (LONG)(scale_factor*width);
	WinROI.top = (ROI.top+ROI.bottom)/2 -  (LONG)(scale_factor*height)/2;
	WinROI.bottom = WinROI.top +  (LONG)(scale_factor*height);

	if(zDelta > 0)
		Invalidate(false);
	else
		Invalidate(true);

	return CMDIChildWnd::OnMouseWheel(nFlags, zDelta, pt);
}
