#pragma once
#include "lssolver.h"
//#include ".././GCToolbox/adjacency_list/graph.h"
#include ".././TPCut/TGraphImage.h"

#include <sys/timeb.h>
#include <time.h>

/*
	Implementation of "Active Contour Without Edges"
	which solves the Energy Minimization:
		F(C, c1,c2) = mu*(length(C))^p + v*area(inside C)
		+ \lambda_1 \int_{inside(C)}{|u_0 - c1|^2 dx}  
		+ \lambda_2 \int_{outside(C)}{|u_0 - c2|^2 dx};
	whose level set formulation is
		F(varpi, c1,c2) = mu *\int_{\Omega}{\delta H(\Omega)}
		+ v * \int_{\Omega}H(\Omega)
		+ \lambda_1 \int_{\Omega}{|u_0 - c1|^2 H(\Omega)dx}
		+ \lambda_2 \int_{\Omega}{|u_0 - c2|^2 (1-H(\Omega))dx}
*/

#define PI 3.1415926


class ActWESolver :
	public LSSolver
{
	double epsilon;	// A small value for calculation of Heaviside function
	double mu;
	double v;
	double lambda1, lambda2;
	double c1, c2;
	double prev_c1,prev_c2;
	double p;
	double k;

	CvMat *u0;
// For graph cuts algorithm
	TGraphImage *g;
	CvMat *dev_rows, *dev_cols, *dev_diag45, *dev_diag135; // a binary matrix recording whether the discontinuity is set.	
// For evaluation of performance
	__time64_t timebuffer_before;
	__time64_t timebuffer_after;	
public:
	int iterate_num;
	__time64_t elapsetime_ms;
	__time64_t totol_elapsetime_ms;
public:
	ActWESolver(void);
	~ActWESolver(void);

	double H(double val); // The Heaviside function
	double Delta(double val);	// The derivative of Heaviside function
	void loadParam(CvMat *varpi, CvMat *u0,double mu, double v, 
		double lambda1, double lambda2,
		double p, double epsilon, double dt, double k, bool video, char *videofname);

	void Iterate();						//Iterate one step of level set equation
	void IterateGC();
	void calC();				// Calculate the value of C1, C2 according to 
	double DenoX(int i, int j);
	double DenoY(int i, int j);
	void Extract_Contour(CvMat *disfun, CvMat *targetMat, double threshold);

	void loadContour(int *x, int *y, int num);

	int what_segment(int x, int y);
};
