#include "StdAfx.h"
#include ".\distancefun.h"
#include "math.h"
#include <stdio.h>      // standard input/output
#include <vector>       // stl vector header
#include "../CVCommon.h"

using namespace std;    // saves us typing std:: before vector


DistanceFun::DistanceFun(ROICoder *roicoder):rcoder(roicoder)
{
	disfun = cvCreateMat(rcoder->imageWidth(), rcoder->imageHeight(), CV_32FC1);
	maxvalue = 0;

	CvSeq *seq = rcoder->getContour(rcoder->getCNum() - 1);
	
	n = rcoder->imageWidth();			// width
	p = rcoder->imageHeight();			// height
	D = new double[n*p];
	S = new double[n*p];
	W = NULL;
	start_points = new double[2*seq->total];
	// Initialize the list chain
	for (int i=0; i<seq->total; i++){
		CvPoint *pnt = (CvPoint*)cvGetSeqElem(seq, i);
		start_points[0+2*i] = pnt->x;
		start_points[1+2*i] = pnt->y;
	}
	end_points = NULL;
	H = NULL;//new double[n*p];;
	nb_iter_max = n*p;//100000;//
	nb_start_points = seq->total;
	nb_end_points = 0;
	heap_pool = NULL;
}

DistanceFun::~DistanceFun(void)
{
	delete[] D;
	delete[] S;
	delete[] start_points;
}

void DistanceFun::CalcuDisFun(int simple=0)
{
	if(!simple)
		perform_front_propagation_2d(NULL);
	int width = rcoder->imageWidth();
	int height = rcoder->imageHeight();	

	IplImage *rgmask = rcoder->GetRegionMask();
	if(!simple){
		for(int i=0; i<width; i++)
			for(int j=0; j<height; j++){	
				double val = D_(i,j);

				if(GetSingleIntValue(rgmask, i, j) == 255)
					cvmSet(disfun, i,j,-val);
				else
					cvmSet(disfun, i,j,val);			
			}
	}else{
		for(int i=0; i<width; i++)
			for(int j=0; j<height; j++){	
				if(GetSingleIntValue(rgmask, i, j) == 255)
					cvmSet(disfun, i,j,-1);
				else
					cvmSet(disfun, i,j,1);			
			}
	}
	maxvalue = 0.5f;

}

float DistanceFun::GetDisFun(int x, int y)
{
	return (float)cvmGet(disfun,x,y);
}

CvMat * DistanceFun::GetDisFun()
{
	return disfun;
}

struct point
{
	point( int ii, int jj )
	{ i = ii; j = jj; }
	int i,j;
};

double *D;
int n;

inline int compare_points(void *x, void *y)
{
	point& a = *( (point*) x );
	point& b = *( (point*) y );
	return cmp( D_(a.i,a.j), D_(b.i,b.j) );
}

bool DistanceFun::end_points_reached(const int i, const int j )
{
	return false;
}

typedef std::vector<point*> point_list;

void DistanceFun::perform_front_propagation_2d(T_callback_intert_node callback_insert_node=NULL)
{
	// create the Fibonacci heap
	struct fibheap* open_heap = fh_makeheap();
	fh_setcmp(open_heap, compare_points);

	double h = 1.0;// /n;	// eliminate '/n' to see the true image distance;

	::D = this->D;
	::n = this->n;
	// initialize points
	for( int i=0; i<n; ++i )
		for( int j=0; j<p; ++j ){
			D_(i,j) = GW_INFINITE;
			S_(i,j) = kFar;
		}

	// record all the points
	heap_pool = new fibheap_el*[n*p]; 
	memset( heap_pool, NULL, n*p*sizeof(fibheap_el*) );

	// inialize open list
	point_list existing_points;
	for( int k=0; k<nb_start_points; ++k )
	{
		int i = (int) start_points_(0,k);
		int j = (int) start_points_(1,k);

		if( D_( i,j )==0 )
			ErrMsgTxt("start_points should not contain duplicates.");

		point* pt = new point( i,j );
		existing_points.push_back( pt );			// for deleting at the end
		heap_pool_(i,j) = fh_insert( open_heap, pt );			// add to heap
		D_( i,j ) = 0;
		S_( i,j ) = kOpen;
	}

	// perform the front propagation
	int num_iter = 0;
	bool stop_iteration = GW_False;
	while( !fh_isempty(open_heap) )//&& num_iter<nb_iter_max&& !stop_iteration
	{
		num_iter++;

		// current point
		point& cur_point = * ((point*) fh_extractmin( open_heap ));
		int i = cur_point.i;
		int j = cur_point.j;
		heap_pool_(i,j) = NULL;
		S_(i,j) = kDead;
		stop_iteration = end_points_reached(i,j);

//		CHECK_HEAP;

		// recurse on each neighbor
		int nei_i[4] = {i+1,i,i-1,i};
		int nei_j[4] = {j,j+1,j,j-1};
		for( int k=0; k<4; ++k )
		{
			int ii = nei_i[k];
			int jj = nei_j[k];
			bool bInsert = true;
			if( callback_insert_node!=NULL )
				bInsert = callback_insert_node(i,j,ii,jj);
			if( ii>=0 && jj>=0 && ii<n && jj<p && bInsert )
			{
				double P = h/W_(ii,jj);
				// compute its neighboring values
				double a1 = GW_INFINITE;
				if( ii<n-1 )
				{
					bool bParticipate = true;
					if( callback_insert_node!=NULL )
						bParticipate = callback_insert_node(ii,jj,ii+1,jj);
					if( bParticipate )
						a1 = D_(ii+1,jj);
				}
				if( ii>0 )
				{
					bool bParticipate = true;
					if( callback_insert_node!=NULL )
						bParticipate = callback_insert_node(ii,jj,ii-1,jj);
					if( bParticipate )
						a1 = GW_MIN( a1, D_(ii-1,jj) );
				}
				double a2 = GW_INFINITE;
				if( jj<p-1 )
				{
					bool bParticipate = true;
					if( callback_insert_node!=NULL )
						bParticipate = callback_insert_node(ii,jj,ii,jj+1);
					if( bParticipate )
						a2 = D_(ii,jj+1);
				}
				if( jj>0 )
				{

					bool bParticipate = true;
					if( callback_insert_node!=NULL )
						bParticipate = callback_insert_node(ii,jj,ii,jj-1);
					if( bParticipate )
						a2 = GW_MIN( a2, D_(ii,jj-1) );
				}
				if( a1>a2 )	// swap so that a1<a2
				{
					double tmp = a1; a1 = a2; a2 = tmp;
				}
				// update its distance
				// now the equation is   (a-a1)^2+(a-a2)^2 = P, with a >= a2 >= a1.
				double A1 = 0;
				if( P*P > (a2-a1)*(a2-a1) )
				{
					double delta = 2*P*P-(a2-a1)*(a2-a1);
					A1 = (a1+a2+sqrt(delta))/2.0;
				}
				else
					A1 = a1 + P;
				if( ((int) S_(ii,jj)) == kDead )
				{
					// check if action has change. Should not happen for FM
					// if( A1<D_(ii,jj) )
					//	mexWarnMsgTxt("The update is not monotone");
#if 1
					if( A1<D_(ii,jj) )	// should not happen for FM
						D_(ii,jj) = A1;
#endif
				}
				else if( ((int) S_(ii,jj)) == kOpen )
				{
					// check if action has change.
					if( A1<D_(ii,jj) )
					{
						D_(ii,jj) = A1;
						// Modify the value in the heap
						fibheap_el* cur_el = heap_pool_(ii,jj);
						if( cur_el!=NULL )
							fh_replacedata( open_heap, cur_el, cur_el->fhe_data );	// use same data for update
						else
							ErrMsgTxt("Error in heap pool allocation."); 
					}
				}
				else if( ((int) S_(ii,jj)) == kFar )
				{
					if( D_(ii,jj)!=GW_INFINITE )
						ErrMsgTxt("Distance must be initialized to Inf");  
					S_(ii,jj) = kOpen;
					// distance must have change.
					D_(ii,jj) = A1;
					// add to open list
					point* pt = new point(ii,jj);
					existing_points.push_back( pt );
					heap_pool_(ii,jj) = fh_insert( open_heap, pt );			// add to heap	
				}
				else 
					ErrMsgTxt("Unknown state."); 
			}	// end switch
		}		// end for
	}			// end while


	// free heap
	fh_deleteheap(open_heap);
	// free point pool
	for( point_list::iterator it = existing_points.begin(); it!=existing_points.end(); ++it )
		GW_DELETE( *it );
	// free fibheap pool
	GW_DELETEARRAY(heap_pool);

}

void DistanceFun::ErrMsgTxt(char *)
{

}

