#include ".\sdistmap.h"

int SDistMap::getInvPointIndex(int x, int y)
{
	contour_list_ptr lpt = m_contourLast;
	point pnt;
	int index = 0;
	if(lpt != NULL) {
		vector<point>::iterator itr = lpt->pnts.begin();
		for( ; itr != lpt->pnts.end(); itr++, index++) {
			pnt = *itr;
			if(pnt.x == x && pnt.y == y) {
				return index;
			}
		}		
	}	

	errfunction("Error in getInvPointIndex.");
	_ASSERT(false);

	return -1;
}

void SDistMap::codeLMap()
{
	int i = m_contourLast->m_numRMin + 1;
	int numpnt = (int)m_contourLast->pnts.size();

	while(i < m_contourLast->m_numRMax){
		int j = m_contourLast->m_numCMin;
		int isInside = 0;
		while(j < m_contourLast->m_numCMax){		
			if(getLMap(i, j) != EDGE){
				if(isInside){
					setLMap(i, j, INNER);
				}					
			}else{// Meet an edge point
				int indx = getInvPointIndex(i, j);

				point pnt1 = m_contourLast->pnts[(indx - 1 + numpnt)%numpnt];
				point pnt2 = m_contourLast->pnts[(indx + 1 + numpnt)%numpnt];
				int signs = (pnt1.x - i)*(pnt2.x - i);
				int dir = 0;
				int left = 0, right = 0;
				if(signs < 0){
					isInside = !isInside;
					j++;
					continue;
				}else if(signs > 0){
					j++;
					continue;
				}else{
					if(pnt1.x != i){
						dir = 1;
						if(pnt1.x < i){
							left = 1;
						}else{
							right = 1;
						}
					}else{
						dir = -1;
						if(pnt2.x < i){
							left = 1;
						}else{
							right = 1;
						}
					}
				}

				while(getLMap(i, j+1) == EDGE){
					indx = (indx + dir + numpnt)%numpnt;					
					_ASSERT(m_contourLast->pnts[indx].x == i);
					j++;
				}

				point node = m_contourLast->pnts[(indx + dir + numpnt)%numpnt];
				if(node.x < i){
					left = 1;
				}else if(node.x > i){
					right = 1;
				}else{
					_ASSERT(false);
				}

				if(left*right != 0){// same direction
					isInside = !isInside;
				}
			}
			j++;
		}
		i++;		
		j = m_contourLast->m_numCMin;
	}
}

void SDistMap::initDMap()
{	
	contour_list_ptr cpt = m_contourFirst;
	point pnt;

	while(!m_actSet.empty())
		m_actSet.pop_front();

	while(cpt != NULL) {
		vector<point>::iterator itr = cpt->pnts.begin();
		for( ; itr!=cpt->pnts.end(); itr++) {
			pnt = *itr;
			if(getDMap(pnt.x, pnt.y) != 0) { // there are intersections
				m_actSet.push_back(pnt);
				setDMap(pnt.x, pnt.y, 0);
			}		
		}
		cpt = cpt->next;
	}
}

void SDistMap::codeDMap()
{
	initDMap();
	while(!m_actSet.empty()) {
		point cur_point = extract_front();
		int x = cur_point.x;
		int y = cur_point.y;
		double cur_dist = getDMap(x, y);
		// recurse on each neighbor
		int nb_x[4] = {x+1, x, x-1, x};
		int nb_y[4] = {y, y+1, y, y-1};
		for( int i=0; i<4; i++ )	{
			int ind_x = nb_x[i];
			int ind_y = nb_y[i];
			if(ind_x >=0 && ind_x < m_width && ind_y >= 0 && ind_y < m_height ) {
				if( !(getDMap(ind_x, ind_y) < INFINITE_D) ) {									
					setDMap(ind_x, ind_y, cur_dist + 1);
					point pnt(ind_x, ind_y);
					m_actSet.push_back(pnt);
				}
			}
		}// End for
	}// End while
}