/*
*/

#include "TGraphImage.h"

namespace NB{
    // Up, Right, Left, Down
	const int nb_x[4] = { -1,  0,  0,  1 };
	const int nb_y[4] = {  0,  1,  -1, 0 };
}

inline bool TGraphImage::is_simple(int x, int y)
{
	/* 8-neighborhood difference in image grid */
	static const int nb_x[9] =    { -1,  -1,  -1,  0,  1,  1,   1,  0, -1 };
	static const int nb_y[9] =    { -1,   0,   1,  1,  1,  0,  -1, -1, -1 };
	static const int abs_xy[9] = {  2,   1,    2,  1,  2,  1,   2,  1,  2 };
	static TLabel prev_lb, cur_lb;
	static int cur_x, cur_y;
	static int label_change, record_idx, record_idx_f, change_num, prev_change_num, foreground_num, background_num;

	if(nb_type == NB_FOUR) {
		change_num = 1;	
		prev_change_num = 0;
		if( x > 0 && y > 0 && x < m_width -1 && y < m_height - 1) {	
			prev_lb = getNode(x + nb_x[0], y + nb_y[0])->t_label;		
			label_change = 0;
			for (int i = 1; i < 9; i++) {
				cur_x = x + nb_x[i];
				cur_y = y + nb_y[i];
				cur_lb = getNode(cur_x, cur_y)->t_label;
				if(cur_lb != prev_lb)  {// a label change occured
					if(label_change > 1)
						return false;
					else 
						label_change++;
					if(change_num == 7) record_idx = i;			
					else
						record_idx = i - 1;
					prev_change_num = change_num;
					change_num = 0;				
				}
				prev_lb = cur_lb;				
				change_num ++;
			}
			if(getNode(x, y)->t_label == FOREGROUND) 
				return (label_change != 0);
			else {
				if(label_change == 0) 
					return false;
				else if(label_change == 2) {
					if(prev_change_num == 1 || prev_change_num == 7)
						return (abs_xy[record_idx] == 1);
					else 
						return true;
				}
				else {
					_ASSERT(false);
					return false;
				}
			}
		}else { // node is on the boundary of the image, always allow topology change
			return false;
		}
	}else { // nb_type == NB_EIGHT

		if( x > 0 && y > 0 && x < m_width -1 && y < m_height - 1) {
			// count the number of foreground
			foreground_num = 0;
			for (int i = 1; i < 9; i++) {
				cur_x = x + nb_x[i];
				cur_y = y + nb_y[i];
				if(getNode(cur_x, cur_y)->t_label == FOREGROUND)	{ 
					record_idx = i;
					foreground_num ++; 
				}else{
					record_idx_f = i;
				}
			}

			if(foreground_num == 0 || foreground_num == 8)		return false; // non-simple
			if(foreground_num == 1)		return true; // simple
			if(foreground_num == 2) {
				if ( getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND ||
					getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND)
					return true;
				else if( (abs_xy[(record_idx-1)%8] == 2 && getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND) ||
					(abs_xy[(record_idx+1)%8] == 2 && getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND)) {
						return true;
				}else {
					return false;
				}
			}
			if(foreground_num == 3) {
				if ( (getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND && 
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND) ||
					(getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND)  ||
					(getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND))
					return true; // consecutive 3
				else if((getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND && 
					abs_xy[(record_idx+2)%8] == 2 &&
					getNode(x + nb_x[(record_idx+3)%8], y + nb_y[(record_idx+3)%8])->t_label == FOREGROUND) ||
					(getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND && 
					abs_xy[(record_idx-2)%8] == 2 &&
					getNode(x + nb_x[(record_idx-3)%8], y + nb_y[(record_idx-3)%8])->t_label == FOREGROUND) ||
					(getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND && 
					abs_xy[(record_idx+1)%8] == 2 &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND) ||
					(getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND && 
					abs_xy[(record_idx-1)%8] == 2 &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND) || 
					(abs_xy[(record_idx-1)%8] == 2 &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND && 
					getNode(x + nb_x[(record_idx-3)%8], y + nb_y[(record_idx-3)%8])->t_label == FOREGROUND) || 
					(abs_xy[(record_idx+1)%8] == 2 &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND && 
					getNode(x + nb_x[(record_idx+3)%8], y + nb_y[(record_idx+3)%8])->t_label == FOREGROUND)) {
						return true;
				}else if( (abs_xy[(record_idx+1)%8] == 2 &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND &&
					abs_xy[(record_idx+3)%8] == 2 &&
					getNode(x + nb_x[(record_idx+4)%8], y + nb_y[(record_idx+4)%8])->t_label == FOREGROUND) || 
					(abs_xy[(record_idx+1)%8] == 2 &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND &&
					abs_xy[(record_idx-1)%8] == 2 &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND) ||
					(abs_xy[(record_idx-1)%8] == 2 &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND &&
					abs_xy[(record_idx-3)%8] == 2 &&
					getNode(x + nb_x[(record_idx-4)%8], y + nb_y[(record_idx-4)%8])->t_label == FOREGROUND)) {
						return true;
				}else{
					return false;
				}

			}

			if(foreground_num == 4) {
				if( (getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+3)%8], y + nb_y[(record_idx+3)%8])->t_label == FOREGROUND) ||
					(getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND ) ||
					(getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND) ||
					(getNode(x + nb_x[(record_idx-3)%8], y + nb_y[(record_idx-3)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND)){
						return true;
				}else if( ( abs_xy[(record_idx+1)%8] == 2 &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+3)%8], y + nb_y[(record_idx+3)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+4)%8], y + nb_y[(record_idx+4)%8])->t_label == FOREGROUND) ||
					( abs_xy[(record_idx-1)%8] == 2 &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-3)%8], y + nb_y[(record_idx-3)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-4)%8], y + nb_y[(record_idx-4)%8])->t_label == FOREGROUND) || 
					( getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND &&
					abs_xy[(record_idx+2)%8] == 2 &&
					getNode(x + nb_x[(record_idx+3)%8], y + nb_y[(record_idx+3)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+4)%8], y + nb_y[(record_idx+4)%8])->t_label == FOREGROUND) ||
					( getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND &&
					abs_xy[(record_idx-2)%8] == 2 &&
					getNode(x + nb_x[(record_idx-3)%8], y + nb_y[(record_idx-3)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-4)%8], y + nb_y[(record_idx-4)%8])->t_label == FOREGROUND) ||
					( getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND &&
					abs_xy[(record_idx+3)%8] == 2 &&
					getNode(x + nb_x[(record_idx+4)%8], y + nb_y[(record_idx+4)%8])->t_label == FOREGROUND) ||
					( getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND &&
					abs_xy[(record_idx-3)%8] == 2 &&
					getNode(x + nb_x[(record_idx-4)%8], y + nb_y[(record_idx-4)%8])->t_label == FOREGROUND) ||
					( abs_xy[(record_idx+1)%8] == 2 &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND &&					
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND) ||
					( abs_xy[(record_idx-1)%8] == 2 &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND &&					
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND) ||
					( abs_xy[(record_idx+1)%8] == 2 &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+3)%8], y + nb_y[(record_idx+3)%8])->t_label == FOREGROUND &&					
					getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND) ||
					( abs_xy[(record_idx-1)%8] == 2 &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-3)%8], y + nb_y[(record_idx-3)%8])->t_label == FOREGROUND &&					
					getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND) ||
					( getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND &&	
					abs_xy[(record_idx+2)%8] == 2 &&					 				
					getNode(x + nb_x[(record_idx+3)%8], y + nb_y[(record_idx+3)%8])->t_label == FOREGROUND) ||
					( getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND &&	
					abs_xy[(record_idx-2)%8] == 2 &&					 				
					getNode(x + nb_x[(record_idx-3)%8], y + nb_y[(record_idx-3)%8])->t_label == FOREGROUND)) {
						return true;
				}else if(( abs_xy[(record_idx+1)%8] == 2 &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND &&
					abs_xy[(record_idx+3)%8] == 2 &&
					getNode(x + nb_x[(record_idx+4)%8], y + nb_y[(record_idx+4)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+5)%8], y + nb_y[(record_idx+5)%8])->t_label == FOREGROUND) ||
					( abs_xy[(record_idx-1)%8] == 2 &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND &&
					abs_xy[(record_idx-3)%8] == 2 &&
					getNode(x + nb_x[(record_idx-4)%8], y + nb_y[(record_idx-4)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-5)%8], y + nb_y[(record_idx-5)%8])->t_label == FOREGROUND) ||
					( abs_xy[(record_idx+1)%8] == 2 &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND &&
					abs_xy[(record_idx+3)%8] == 2 &&
					getNode(x + nb_x[(record_idx+4)%8], y + nb_y[(record_idx+4)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND) ||
					( abs_xy[(record_idx-1)%8] == 2 &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND &&
					abs_xy[(record_idx-3)%8] == 2 &&
					getNode(x + nb_x[(record_idx-4)%8], y + nb_y[(record_idx-4)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND) ||
					( abs_xy[(record_idx+1)%8] == 2 &&
					abs_xy[(record_idx-1)%8] == 2 &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND &&					
					getNode(x + nb_x[(record_idx+3)%8], y + nb_y[(record_idx+3)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND) ||
					( abs_xy[(record_idx+1)%8] == 2 &&
					abs_xy[(record_idx-1)%8] == 2 &&
					getNode(x + nb_x[(record_idx-2)%8], y + nb_y[(record_idx-2)%8])->t_label == FOREGROUND &&					
					getNode(x + nb_x[(record_idx-3)%8], y + nb_y[(record_idx-3)%8])->t_label == FOREGROUND &&
					getNode(x + nb_x[(record_idx+2)%8], y + nb_y[(record_idx+2)%8])->t_label == FOREGROUND) ||
					( getNode(x + nb_x[(record_idx+1)%8], y + nb_y[(record_idx+1)%8])->t_label == FOREGROUND &&					
					abs_xy[(record_idx+2)%8] == 2 &&
					getNode(x + nb_x[(record_idx+3)%8], y + nb_y[(record_idx+3)%8])->t_label == FOREGROUND &&					
					abs_xy[(record_idx+4)%8] == 2 &&					
					getNode(x + nb_x[(record_idx+5)%8], y + nb_y[(record_idx+5)%8])->t_label == FOREGROUND) ||
					( getNode(x + nb_x[(record_idx-1)%8], y + nb_y[(record_idx-1)%8])->t_label == FOREGROUND &&					
					abs_xy[(record_idx-2)%8] == 2 &&
					getNode(x + nb_x[(record_idx-3)%8], y + nb_y[(record_idx-3)%8])->t_label == FOREGROUND &&					
					abs_xy[(record_idx-4)%8] == 2 &&					
					getNode(x + nb_x[(record_idx-5)%8], y + nb_y[(record_idx-5)%8])->t_label == FOREGROUND)){
						return true;
				}else{
					return false;
				}
			}

			if(foreground_num == 5) {
				if((getNode(x + nb_x[(record_idx_f+1)%8], y + nb_y[(record_idx_f+1)%8])->t_label == BACKGROUND &&
					getNode(x + nb_x[(record_idx_f+2)%8], y + nb_y[(record_idx_f+2)%8])->t_label == BACKGROUND) ||
					(getNode(x + nb_x[(record_idx_f-1)%8], y + nb_y[(record_idx_f-1)%8])->t_label == BACKGROUND &&
					getNode(x + nb_x[(record_idx_f-2)%8], y + nb_y[(record_idx_f-2)%8])->t_label == BACKGROUND) ||
					(getNode(x + nb_x[(record_idx_f-1)%8], y + nb_y[(record_idx_f-1)%8])->t_label == BACKGROUND &&
					getNode(x + nb_x[(record_idx_f+1)%8], y + nb_y[(record_idx_f+1)%8])->t_label == BACKGROUND)) {
						return true;
				}else{
					background_num = 0;
					for (int i = 1; i < 9; i++) {
						cur_x = x + nb_x[i];
						cur_y = y + nb_y[i];
						if(getNode(cur_x, cur_y)->t_label == BACKGROUND)	{ 
							if(abs_xy[i] == 2)
								background_num ++;
						}
					}

					if(background_num == 2)
						return true;
					else
						return false;
				}
			}

			if(foreground_num == 6) {
				if(getNode(x + nb_x[(record_idx_f+1)%8], y + nb_y[(record_idx_f+1)%8])->t_label == BACKGROUND ||
					getNode(x + nb_x[(record_idx_f-1)%8], y + nb_y[(record_idx_f-1)%8])->t_label == BACKGROUND)
					return true;
				else
					return false;
			}

			if(foreground_num == 7) {
				if(abs_xy[record_idx_f] == 2) 
					return false;
				else 
					return true;					
			}
		}else {//on the boundary
			return false;
		}


		return true;
	}
}

inline TGraphImage::t_node* TGraphImage::next_active(TNODE_LPQUEUE *act_set)
{
	if(act_set->is_empty())
		return NULL;

	t_node *nd = act_set->front();
	if(nd) {
		act_set->pop();
		nd->node_ptr->node_state = &STATE_FREE;
	}
	return nd;
}

/* pop out the one with the larger top() -- to select the one farther to the boundary first */
inline TGraphImage::t_node* TGraphImage::next_active(TNODE_LPQUEUE *act_set1, TNODE_LPQUEUE *act_set2)
{
	t_node *nd1, *nd2;

	if(act_set1->is_empty())	nd1 = NULL;
	else nd1 = act_set1->front();

	if(act_set2->is_empty())	nd2 = NULL;
	else nd2 = act_set2->front();

	if(nd1 == NULL && nd2 == NULL) {
		return NULL;
	}else if(nd1 == NULL) {		
		act_set2->pop();
		nd2->node_ptr->node_state = &STATE_FREE;
		return nd2;
	}else if(nd2 == NULL) {
		act_set1->pop();
		nd1->node_ptr->node_state = &STATE_FREE;
		return nd1;
	}else if(nd1->node_ptr->DIST > nd2->node_ptr->DIST) {		
		act_set1->pop();		
		nd1->node_ptr->node_state = &STATE_FREE;
		return nd1;
	}else {
		act_set2->pop();	
		nd2->node_ptr->node_state = &STATE_FREE;
		return nd2;
	}
}

inline void TGraphImage::set_active(TNODE_LPQUEUE *act_set, t_node *nd)
{
	act_set->push_back(nd);
}

inline void TGraphImage::set_passive(TNODE_LPQUEUE *pas_set, t_node *nd)
{
	pas_set->push_back(nd);
}

void TGraphImage::exchange_set(TNODE_LPQUEUE *& set1, TNODE_LPQUEUE *& set2)
{	
	// exchange pointer
	TNODE_LPQUEUE *tmp_set = set1;
	set1 = set2;
	set2 = tmp_set;
	// exchange tag
	NodeState tmp_stat = set1->get_tag();
	set1->set_tag(set2->get_tag());
	set2->set_tag(tmp_stat);
}

void TGraphImage::empty_set(TNODE_LPQUEUE *set)
{
	while(!set->is_empty())
		set->pop();
}

void TGraphImage::delete_node(TNODE_LPQUEUE *set, t_node *nd)
{
	set->delete_item(nd);
}

void TGraphImage::transform(t_node *node1, t_node *node2, arc *arc_ptr, bool is_sink)
{
	CapType min_cap, tmp_cap, retract_cap;
	t_node *nd = node1;
	
	if( !is_sink ) { // node1 is a source tree, arc_ptr = (n1, n2) node2 should be a free node
		min_cap = arc_ptr->r_cap;
		while(nd->parent_node != TERMINAL) {
			tmp_cap = nd->arc_ptr[nd->parc_ind]->sister->r_cap;
			if(min_cap > tmp_cap) min_cap = tmp_cap;
			nd = nd->parent_node;
		}
		/* capacity of terminal */
		tmp_cap = nd->node_ptr->tr_cap;
		if( min_cap > tmp_cap ) min_cap = tmp_cap;

		_ASSERT(min_cap > 0);

		retract_cap = min_cap / 2.0;
		/* now transform node2 to be terminal */
		node2->node_ptr->tr_cap = retract_cap; 

		nd->node_ptr->tr_cap -= retract_cap;

		arc_ptr->r_cap -= retract_cap;
		arc_ptr->sister->r_cap += retract_cap;

		nd = node1;
		while(nd->parent_node != TERMINAL) {
			nd->arc_ptr[nd->parc_ind]->sister->r_cap -= retract_cap;
			nd->arc_ptr[nd->parc_ind]->r_cap += retract_cap;
			nd = nd->parent_node;
		}
	}else {// node1 is a sink tree, arc_ptr = (n2, n1)
		min_cap = arc_ptr->r_cap;
		while(nd->parent_node != TERMINAL) {
			tmp_cap = nd->arc_ptr[nd->parc_ind]->r_cap;
			if(min_cap > tmp_cap) min_cap = tmp_cap;
			nd = nd->parent_node;
		}
		/* capacity of terminal */
		tmp_cap = - nd->node_ptr->tr_cap;
		if( min_cap > tmp_cap ) min_cap = tmp_cap;

		_ASSERT(min_cap > 0);

		retract_cap = min_cap / 2.0;
		/* now transform node2 to be terminal */
		node2->node_ptr->tr_cap = -retract_cap; 

		nd->node_ptr->tr_cap += retract_cap;

		arc_ptr->r_cap -= retract_cap;
		arc_ptr->sister->r_cap += retract_cap;

		nd = node1;
		while(nd->parent_node != TERMINAL) {
			nd->arc_ptr[nd->parc_ind]->r_cap -= retract_cap;
			nd->arc_ptr[nd->parc_ind]->sister->r_cap += retract_cap;
			nd = nd->parent_node;
		}
	}
	node2->parent_node = TERMINAL;
}

void TGraphImage::maxflow_init()
{
	empty_set( act_Sf );		empty_set( act_Sb );	empty_set( act_Tf );		empty_set( act_Tb );
	empty_set( pas_Sf );	empty_set( pas_Sb );	empty_set( pas_Tf );	empty_set( pas_Tb );

	/* encode topology label and distance*/
	for (int i = 0; i < m_width; i++) {
		for (int j = 0; j < m_height; j++) {
			t_node *nd = getNode(i, j);
			if( sdmap->what_Label(i, j) != SDistMap::OUTER )
				nd->t_label = FOREGROUND;
			nd->node_ptr->DIST = (int)sdmap->what_Distance(i, j);
		}
	}

	for (int i = 0; i < m_width; i++) {
		for (int j = 0; j < m_height; j++) {
			t_node *nd = getNode(i, j);
			if(nd->node_ptr->tr_cap > 0) { // in the source(S) tree
				nd->node_ptr->tree_type = SOURCE;
				nd->parent_node = TERMINAL;
				if(getTLabel(nd) == BACKGROUND) {					
					set_active(act_Sb, nd);
				}else {					
					set_active(act_Sf, nd);
				}
			}else if(nd->node_ptr->tr_cap < 0){ // in the sink(T) tree
				nd->node_ptr->tree_type = SINK;
				nd->parent_node = TERMINAL;
				if(getTLabel(nd) == BACKGROUND) {					
					set_active(act_Tb, nd);
				}else {					
					set_active(act_Tf, nd);
				}
			}else {	// a free node
				nd->parent_node = NULL;
				nd->node_ptr->tree_type = NONE;
			}
		}
	}

	act_Sf->begin();		act_Sb->begin();	act_Tf->begin();	act_Tb->begin();
	pas_Sf->begin();	pas_Sb->begin();	pas_Tf->begin();	pas_Tb->begin();
}

Graph::FlowType TGraphImage::tp_maxflow()
{
	t_node *n_i, *n_j, *current_node;
	bool found_feasible;
	arc *cur_arc; // record the arc (src_node, sink_node) when augment is to be call
    

	while(!act_Tb->is_empty() || !act_Tf->is_empty() || !act_Sf->is_empty() || !act_Sb->is_empty()) {
			/* Phase 1.	Boundary Competition Flow */
			while(!act_Tb->is_empty() || !act_Tf->is_empty() || !act_Sf->is_empty() || !act_Sb->is_empty()) {
				/* 1.a	Act_Sf  <->  Act_Tb */
				current_node = NULL;
				while (1) {
					found_feasible = false;
					if(n_i = current_node) {
						if(!n_i->parent_node) n_i = NULL;
					}
					if(!n_i) {
						if( !(n_i = next_active(act_Sf, act_Tb)) ) break;
					}

					/* growth */		
					if(n_i->node_ptr->tree_type == SOURCE) { // n_i in Act_Sf				
						if(cur_arc = source_boundary_competition_flow(n_i, n_j, act_Sf, act_Sb, pas_Sb))
							found_feasible = true;
					}
					else if(n_i->node_ptr->tree_type == SINK) { // n_i in Act_Tb				
						if(cur_arc = sink_boundary_competition_flow(n_i, n_j, act_Tb, act_Tf, pas_Tf))
							found_feasible = true;
					}
					else { // n_j's tree_type is NONE
						_ASSERT(false);
					}

					if( found_feasible ) {
						current_node = n_i;
						/* augment */
						if( n_i->node_ptr->tree_type == SOURCE)
							augment(n_i, n_j, cur_arc);
						else
							augment(n_j, n_i, cur_arc);

						/* adoption */
						while( !orphan_set.empty() ) {
							t_node *orphan_nd = orphan_set.front();
							orphan_set.pop();
							if(orphan_nd->node_ptr->tree_type == SOURCE) 
								process_source_orphan( orphan_nd );
							else if(orphan_nd->node_ptr->tree_type == SINK)
								process_sink_orphan( orphan_nd );
							else
								_ASSERT(false);
						}
					}else { // n_i is passive now
						(n_i->node_ptr->tree_type == SOURCE) ? pas_Sf->push_back(n_i) : pas_Tb->push_back(n_i);
						current_node = NULL;
					}
				}

				/* 1.b	Act_Sb <-> Act_Tf */
				current_node = NULL;
				while (1) {
					found_feasible = false;
					if(n_i = current_node) {
						if(!n_i->parent_node) n_i = NULL;
					}
					if(!n_i) {
						if( !(n_i = next_active(act_Sb, act_Tf)) ) break;
					}

					/* growth */		
					if(n_i->node_ptr->tree_type == SOURCE) { // n_i in Act_Sb			
						if(cur_arc = source_boundary_competition_flow(n_i, n_j, act_Sb, act_Sf, pas_Sf))
							found_feasible = true;
					}
					else if(n_i->node_ptr->tree_type == SINK) { // n_i in Act_Tf				
						if(cur_arc = sink_boundary_competition_flow(n_i, n_j, act_Tf, act_Tb, pas_Tb))
							found_feasible = true;
					}
					else { // n_j's tree_type is NONE
						_ASSERT(false);
					}

					if( found_feasible ) {
						current_node = n_i;
						/* augment */
						if( n_i->node_ptr->tree_type == SOURCE )
							augment(n_i, n_j, cur_arc);
						else
							augment(n_j, n_i, cur_arc);

						/* adoption */
						while( !orphan_set.empty() )	{
							t_node *orphan_nd = orphan_set.front();
							orphan_set.pop();
							if(orphan_nd->node_ptr->tree_type == SOURCE) 
								process_source_orphan( orphan_nd );
							else if(orphan_nd->node_ptr->tree_type == SINK)
								process_sink_orphan( orphan_nd );
							else
								_ASSERT(false);
						}
					}else { // n_i is passive now
						if(n_i->node_ptr->tree_type == SOURCE) {
							if(is_simple(n_i->pos_x, n_i->pos_y)) {
								act_Sf->push_back(n_i);
								n_i->t_label = FOREGROUND;
								break;
							}else
								pas_Sb->push_back(n_i);
						}else{
							pas_Tf->push_back(n_i);
						}
						current_node = NULL;
					}
				}
			}// end while(!act_Tb.empty() || !act_Tf.empty() || !act_Sf.empty() || !act_Sb.empty())

			/* Phase 2.	Intra-region Topology Preserving Flow */
			exchange_set(act_Tb, pas_Tb);
			exchange_set(act_Sf, pas_Sf);
			exchange_set(act_Sb, pas_Sb);
			exchange_set(act_Tf, pas_Tf);
			/* 2.a	Act_Sf <-> Act_Tf */
			current_node = NULL;
			while( !act_Tf->is_empty() || current_node != NULL) {
				found_feasible = false;
				if(n_i = current_node) {
					if(!n_i->parent_node) n_i = NULL;
				}
				if(!n_i)  {
					if(!(n_i = next_active(act_Tf)) ) break;	
				}

				/* growth */	
				if (cur_arc = foreground_intra_region_flow(n_i, n_j))
					found_feasible = true;

				if( found_feasible ) {
					current_node = n_i;
					/* augment */
					if( n_i->node_ptr->tree_type == SOURCE)
						augment(n_i, n_j, cur_arc);
					else
						augment(n_j, n_i, cur_arc);

					/* adoption */
					while( !orphan_set.empty() )	{
						t_node *orphan_nd = orphan_set.front();
						orphan_set.pop();
						if(orphan_nd->node_ptr->tree_type == SOURCE)
							process_source_orphan( orphan_nd );				
						else
							process_sink_orphan( orphan_nd );
					}
				}else { // n_i is passive now
					pas_Tf->push_back(n_i);
					current_node = NULL;
				}       
			}// end while( !act_Tf->is_empty() || current_node != NULL)

			/* 2.b	Act_Sb <-> Act_Tb */
			current_node = NULL;
			while( !act_Sb->is_empty() || current_node != NULL) {
				found_feasible = false;
				if(n_i = current_node) {
					if(!n_i->parent_node) n_i = NULL;
				}
				if(!n_i) {
					if(!(n_i = next_active(act_Sb))) break;	
				}

				/* growth */	
				if (cur_arc = background_intra_region_flow(n_i, n_j))
					found_feasible = true;

				if( found_feasible ) {
					current_node = n_i;
					/* augment */
					augment(n_i, n_j, cur_arc);

					/* adoption */
					while( !orphan_set.empty() ) {
						t_node *orphan_nd = orphan_set.front();
						orphan_set.pop();
						if(orphan_nd->node_ptr->tree_type == SOURCE)
							process_source_orphan( orphan_nd );
						else
							process_sink_orphan( orphan_nd );
					}
				}else { // n_i is passive now
					pas_Sb->push_back(n_i);
					current_node = NULL;
				}       
			}// end while( !act_Sb->is_empty() || current_node != NULL)
			
			// label correction
			//_ASSERT(act_Sb->is_empty() && act_Tf->is_empty());
			empty_set(act_Sb);
			empty_set(act_Tf);
			empty_set(act_Tb);
			empty_set(act_Sf);
			empty_set(pas_Tb);
			empty_set(pas_Tf);
			empty_set(pas_Sb);
			empty_set(pas_Sf);

				while( !foreground_simple_set.empty() ) {
				n_i = foreground_simple_set.front();
				if(is_simple(n_i->pos_x, n_i->pos_y)) {
					n_i->t_label = BACKGROUND;
					if(n_i->parent_node != NULL) {// not free node
						if(n_i->node_ptr->tr_cap == 0) { // not directed linked to the terminals
							transform(n_i->parent_node, n_i, n_i->arc_ptr[n_i->parc_ind]->sister, 1);
						}					
						act_Tb->push_back(n_i);
					}
				}
				foreground_simple_set.pop();
				n_i->is_candidate = 0;
			}
		while( !background_simple_set.empty() ) {
				n_i = background_simple_set.front();
				if(is_simple(n_i->pos_x, n_i->pos_y)) {
					n_i->t_label = FOREGROUND;
					if(n_i->parent_node != NULL) {// not free node
						if(n_i->node_ptr->tr_cap == 0) { // not directed linked to the terminals
							transform(n_i->parent_node, n_i, n_i->arc_ptr[n_i->parc_ind], 0);
						}
						act_Sf->push_back(n_i);
					}
				}
				background_simple_set.pop();
				n_i->is_candidate = 0;
			}
	}

	/* Phase 3. Inter-region Topology Competition Flow */

	return flow;
}

Graph::arc* TGraphImage::source_boundary_competition_flow( 
	t_node *n_i,  t_node* &n_output,  
	TNODE_LPQUEUE *act_Si,  
	TNODE_LPQUEUE *act_So,  
	TNODE_LPQUEUE *pas_So )
{
	int x_j, y_j;
	t_node *n_j;

	for( int k = 0; k < 4; k ++) { // for each of 4-neighbor i in N(x_i)
		x_j = n_i->pos_x + NB::nb_x[k];
		y_j = n_i->pos_y + NB::nb_y[k];
		if( IN_IMAGE(x_j, y_j) ) {
			if( n_i->arc_ptr[k]->r_cap > 0 ) { // cap_arc(n_i, n_j) > 0

				n_j = getNode(x_j, y_j);

				if( n_i->t_label != n_j->t_label ) { // n_j has a different t_label: BACKGROUND
					if( n_j->parent_node == NULL ) { // n_j is a free node
						if( is_simple(n_j->pos_x, n_j->pos_y) ) { // recruit n_j to Act_Sf
							n_j->node_ptr->tree_type = SOURCE;
							n_j->t_label = n_i->t_label;
							n_j->node_ptr->DIST = n_i->node_ptr->DIST - 1;
							n_j->parent_node = n_i;
							n_j->parc_ind = 3 - k;
							act_Si->push_back( n_j );									
						}else { // transform n_j
							transform(n_i, n_j, n_i->arc_ptr[k], false);
							n_j->node_ptr->tree_type = SOURCE;
							act_So->push_back(n_j);
						}
					}else if(n_j->node_ptr->tree_type == SOURCE) {
						if( is_simple(n_j->pos_x, n_j->pos_y) && 
							is_simple(n_i->pos_x, n_i->pos_y) &&
							n_i->t_label == FOREGROUND ) { // recruit n_j
							n_j->t_label = n_i->t_label;												
							if(ND_STATE(n_j) == ACTIVE) {
								act_So->delete_item(n_j);								
							}else if(ND_STATE(n_j) == PASSIVE) {
								pas_So->delete_item(n_j);								
							}else {
								_ASSERT(false);
							}				
							n_j->node_ptr->DIST = n_i->node_ptr->DIST - 1;							
							act_Si->push_back(n_j);
						}
					}else { // n_j's tree type is SINK, and it is in Act_Tb, find a feasible flow
						n_output = n_j;
						return n_i->arc_ptr[k];						
					}
				}
				else { //n_i->t_label == n_j->t_label
					if( n_j->parent_node == NULL ) { // free node
						n_j->node_ptr->tree_type = SOURCE;
						n_j->node_ptr->DIST = n_i->node_ptr->DIST - 1;
						n_j->parent_node = n_i;
						n_j->parc_ind = 3 - k;
						act_Si->push_back( n_j );		
					}
				}
			}// end if( n_i->arc_ptr[k] > 0 )
		}// end if( IN_IMAGE(x, y) )
	}// end for( int k = 0; k < 4; k ++)
	return NULL;
}


Graph::arc* TGraphImage::sink_boundary_competition_flow( 
	t_node *n_i,  t_node* &n_output,  
	TNODE_LPQUEUE *act_Ti,  
	TNODE_LPQUEUE *act_To,  
	TNODE_LPQUEUE *pas_To  )
{
	int x_j, y_j;
	t_node *n_j;

	for( int k = 0; k < 4; k ++) { // for each of 4-neighbor i in N(x_i)
		x_j = n_i->pos_x + NB::nb_x[k];
		y_j = n_i->pos_y + NB::nb_y[k];
		if( IN_IMAGE(x_j, y_j) ) {
			if( n_i->arc_ptr[k]->sister->r_cap > 0 ) {// cap_arc(n_j, n_i) > 0

				n_j = getNode(x_j, y_j);

				if( n_i->t_label != n_j->t_label ) { // n_j has a different t_label: FOREGROUND
					if( n_j->parent_node == NULL ) { // n_j is a free node
						if( is_simple(n_j->pos_x, n_j->pos_y) ) { // recruit n_j to Act_Tb
							n_j->node_ptr->tree_type = SINK;
							n_j->t_label = n_i->t_label;
							n_j->node_ptr->DIST = n_i->node_ptr->DIST - 1;
							n_j->parent_node = n_i;
							n_j->parc_ind = 3 - k;
							act_Ti->push_back( n_j );									
						}else { // transform n_j
							transform(n_i, n_j, n_i->arc_ptr[k]->sister, true);
							n_j->node_ptr->tree_type = SINK;
							act_To->push_back(n_j);
						}
					}else if(n_j->node_ptr->tree_type == SINK) { 
						if( is_simple(n_j->pos_x, n_j->pos_y) && 
							is_simple(n_i->pos_x, n_i->pos_y) && 
							n_i->t_label == BACKGROUND ) { // recruit n_j
							n_j->t_label = n_i->t_label;							
							if(ND_STATE(n_j) == ACTIVE) { 
								act_To->delete_item(n_j);								
							}else if(ND_STATE(n_j) == PASSIVE) {
								pas_To->delete_item(n_j);								
							}else {
								_ASSERT(false);
							}									
							n_j->node_ptr->DIST = n_i->node_ptr->DIST - 1;
							act_Ti->push_back(n_j);
						}
					}else { // n_j's tree type is SOURCE, and it is in Act_Sf, found a feasible flow
						n_output = n_j;
						return n_i->arc_ptr[k]->sister;					
					}
				}
				else { //n_i->t_label == n_j->t_label
					if( n_j->parent_node == NULL ) { // free node
						n_j->node_ptr->tree_type = SINK;
						n_j->node_ptr->DIST = n_i->node_ptr->DIST - 1;
						n_j->parent_node = n_i;
						n_j->parc_ind = 3 - k;
						act_Ti->push_back( n_j );		
					}
				}
			}// end if( n_i->arc_ptr[k] > 0 )
		}// end if( IN_IMAGE(x, y) )
	}// end for( int k = 0; k < 4; k ++)
	return NULL;
}

void TGraphImage::exclude_node(t_node *nd_in, bool is_sink)
{
	if(nd_in->parent_node != TERMINAL) {
		CapType min_cap, tmp_cap, retract_cap;
		t_node *nd = nd_in;

		if( !is_sink ) { // node1 is a source tree, arc_ptr = (n1, n2) node2 should be a free node
			min_cap = 10e10 ;
			while(nd->parent_node != TERMINAL) {
				tmp_cap = nd->arc_ptr[nd->parc_ind]->sister->r_cap;
				if(min_cap > tmp_cap) min_cap = tmp_cap;
				nd = nd->parent_node;
			}
			/* capacity of terminal */
			tmp_cap = nd->node_ptr->tr_cap;
			if( min_cap > tmp_cap ) min_cap = tmp_cap;

			_ASSERT(min_cap > 0);

			retract_cap = min_cap / 2.0;
			/* now transform nd_in to be terminal */
			nd_in->node_ptr->tr_cap = retract_cap; 

			nd->node_ptr->tr_cap -= retract_cap;

			nd = nd_in;
			while(nd->parent_node != TERMINAL) {
				nd->arc_ptr[nd->parc_ind]->sister->r_cap -= retract_cap;
				nd->arc_ptr[nd->parc_ind]->r_cap += retract_cap;
				nd = nd->parent_node;
			}
		}else {// node1 is a sink tree, arc_ptr = (n2, n1)
			min_cap = 10e10;
			while(nd->parent_node != TERMINAL) {
				tmp_cap = nd->arc_ptr[nd->parc_ind]->r_cap;
				if(min_cap > tmp_cap) min_cap = tmp_cap;
				nd = nd->parent_node;
			}
			/* capacity of terminal */
			tmp_cap = - nd->node_ptr->tr_cap;
			if( min_cap > tmp_cap ) min_cap = tmp_cap;

			_ASSERT(min_cap > 0);

			retract_cap = min_cap / 2.0;
			/* now transform nd_in to be terminal */
			nd_in->node_ptr->tr_cap = -retract_cap; 

			nd->node_ptr->tr_cap += retract_cap;

			nd = nd_in;
			while(nd->parent_node != TERMINAL) {
				nd->arc_ptr[nd->parc_ind]->r_cap -= retract_cap;
				nd->arc_ptr[nd->parc_ind]->sister->r_cap += retract_cap;
				nd = nd->parent_node;
			}
		}
		nd_in->parent_node = TERMINAL;
	}
}

Graph::arc* TGraphImage::foreground_intra_region_flow( 
	t_node *n_i, t_node* &n_output )
{
	int x_j, y_j;
	t_node *n_j;

	for( int k = 0; k < 4; k ++) { // for each of 4-neighbor i in N(x_i)
		x_j = n_i->pos_x + NB::nb_x[k];
		y_j = n_i->pos_y + NB::nb_y[k];
		if( IN_IMAGE(x_j, y_j) ) {
			n_j = getNode(x_j, y_j);
			if( n_i->t_label == n_j->t_label  ) {// cap_arc(n_j, n_i) > 0				
				if( n_i->arc_ptr[k]->sister->r_cap > 0 ) {
					if( n_j->parent_node == NULL ) { // free node
						n_j->node_ptr->tree_type = SINK;
						n_j->node_ptr->DIST = n_i->node_ptr->DIST - 1;
						n_j->parent_node = n_i;
						n_j->parc_ind = 3 - k;
						act_Tf->push_back( n_j );
					}
					else if( n_j->node_ptr->tree_type == SOURCE ) { // found a feasible flow
						n_output = n_j;
						return n_i->arc_ptr[k]->sister;			
					}else {} // tree_type == SINK, no action
				}				
			}else { //n_i->t_label != n_j->t_label
					if(is_simple(n_i->pos_x, n_i->pos_y)) { // record n_i
						if(!n_i->is_candidate) {
							foreground_simple_set.push(n_i);						
							n_i->is_candidate = 1;
						}
					}
			}
		}// end of "if( IN_IMAGE(x_j, y_j) )"
	}
	return NULL;
}

Graph::arc* TGraphImage::background_intra_region_flow( 
		  t_node *n_i, t_node* &n_output )
{
	int x_j, y_j;
	t_node *n_j;

	for( int k = 0; k < 4; k ++) { // for each of 4-neighbor i in N(x_i)
		x_j = n_i->pos_x + NB::nb_x[k];
		y_j = n_i->pos_y + NB::nb_y[k];
		if( IN_IMAGE(x_j, y_j) ) {
			n_j = getNode(x_j, y_j);	
			if( n_i->t_label == n_j->t_label  ) {// cap_arc(n_j, n_i) > 0			
				if( n_i->arc_ptr[k]->r_cap > 0 ) {					
					if( n_j->parent_node == NULL ) { // free node
						n_j->node_ptr->tree_type = SOURCE;
						n_j->node_ptr->DIST = n_i->node_ptr->DIST - 1;
						n_j->parent_node = n_i;
						n_j->parc_ind = 3 - k;
						act_Tf->push_back( n_j );
					}
					else if( n_j->node_ptr->tree_type == SINK ) { // found a feasible flow
						n_output = n_j;
						return n_i->arc_ptr[k];			
					}else {} // tree_type == SINK, no action
				}				
			}else { //n_i->t_label != n_j->t_label
					if(is_simple(n_i->pos_x, n_i->pos_y)) { // record n_i
						if(!n_i->is_candidate) {
							background_simple_set.push(n_i);				
							n_i->is_candidate = 1;
						}						
					}
			}
		}// end of "if( IN_IMAGE(x_j, y_j) )"
	}
	return NULL;
}

void TGraphImage::augment(t_node *src_nd, t_node *sink_nd, arc *arc_middle)
{
	CapType cap_bottleneck, cap_tmp;
	t_node *nd, *nd_tmp;

	/* 1. find bottleneck capacity */
	cap_bottleneck = arc_middle->r_cap;

	nd = src_nd;
	for( ; nd->parent_node != TERMINAL; nd = nd->parent_node) {
		cap_tmp = nd->arc_ptr[nd->parc_ind]->sister->r_cap;
		if(cap_bottleneck > cap_tmp)	cap_bottleneck = cap_tmp;
	}
	if(cap_bottleneck > nd->node_ptr->tr_cap)	cap_bottleneck = nd->node_ptr->tr_cap;

	nd = sink_nd;
	for( ; nd->parent_node != TERMINAL; nd = nd->parent_node) {
		cap_tmp = nd->arc_ptr[nd->parc_ind]->r_cap;
		if(cap_bottleneck > cap_tmp)	cap_bottleneck = cap_tmp;
	}
	if(cap_bottleneck > -nd->node_ptr->tr_cap) cap_bottleneck = -nd->node_ptr->tr_cap;


	/* 2. Augmenting */
	arc_middle->r_cap -= cap_bottleneck;
	arc_middle->sister->r_cap += cap_bottleneck;

	/* 2.a - the source tree */
	nd = src_nd;
	for( ; nd->parent_node != TERMINAL; nd = nd_tmp) {
		nd_tmp = nd->parent_node;
		nd->arc_ptr[nd->parc_ind]->sister->r_cap -= cap_bottleneck;
		nd->arc_ptr[nd->parc_ind]->r_cap += cap_bottleneck;
		if( !nd->arc_ptr[nd->parc_ind]->sister->r_cap ) {			
			nd->parent_node = ORPHAN;
			orphan_set.push(nd);
		}
	}
	nd->node_ptr->tr_cap -= cap_bottleneck;
	if( !nd->node_ptr->tr_cap ) {
		nd->parent_node = ORPHAN;
		orphan_set.push(nd);
	}

	/* 2.b - the sink tree */
	nd = sink_nd;
	for( ; nd->parent_node != TERMINAL; nd = nd_tmp) {
		nd_tmp = nd->parent_node;
		nd->arc_ptr[nd->parc_ind]->r_cap -= cap_bottleneck;
		nd->arc_ptr[nd->parc_ind]->sister->r_cap += cap_bottleneck;
		if( !nd->arc_ptr[nd->parc_ind]->r_cap ) {			
			nd->parent_node = ORPHAN;
			orphan_set.push(nd);
		}
	}
	nd->node_ptr->tr_cap += cap_bottleneck;
	if( !nd->node_ptr->tr_cap ) {
		nd->parent_node = ORPHAN;
		orphan_set.push(nd);
	}

	flow += (FlowType)cap_bottleneck;
}

void TGraphImage::process_source_orphan(t_node *n_i)
{
	int x_j, y_j;
	t_node *n_j, *n_tmp;
	bool found_terminal = false;
	int k_min, dist, dist_min = INFINITE_D;
	t_node *nd_min = NULL;
	
	/* try to find a new parent with the same label */
	for(int k = 0; k < 4; k++) {
		x_j = n_i->pos_x + NB::nb_x[k];
		y_j = n_i->pos_y + NB::nb_y[k];
		if( IN_IMAGE(x_j, y_j) ) {
			n_j = getNode(x_j, y_j);
			if( n_j->node_ptr->tree_type == SOURCE && n_j->t_label == n_i->t_label && n_i->arc_ptr[k]->sister->r_cap > 0 )  {				
				dist = 0;
				n_tmp = n_j;
				while(1) {
					if( n_tmp->parent_node == ORPHAN || n_tmp->parent_node == n_i) {						
						dist = INFINITE_D;
						break;
					}else if(n_tmp->parent_node == TERMINAL) {
						found_terminal = true;						
						break;
					}else {
						n_tmp = n_tmp->parent_node;
						dist++;
					}
				}
				// find the shortest to the path to the terminal
				if(dist == 0) { // to reduce iteration number
					n_i->parent_node = n_j;
					n_i->parc_ind = k;
					return;
				}else if(dist < dist_min) { 
					nd_min = n_j;
					k_min = k;
				}
			}			
		}// end if( IN_IMAGE(x_j, y_j) )
	}// end for(int k = 0; k < 4; k++)

	if( found_terminal ) {
		n_i->parent_node = nd_min;
		n_i->parc_ind = k_min;				
		return;
	}

	/* no new legal parent is found, now n_i is free */
	if(ND_STATE(n_i) != FREE) { // it belonged to an active set or passive set
		/* in act_Sf */
		if( ND_STATE(n_i) == ACTIVE ) 
			if(getTLabel(n_i) == FOREGROUND)
				delete_node(act_Sf, n_i);
			else
				delete_node(act_Sb, n_i);
		else 
			if(getTLabel(n_i) == FOREGROUND)
				delete_node(pas_Sf, n_i);			
			else
				delete_node(pas_Sb, n_i);
	}
	n_i->parent_node = NULL;	
	n_i->node_ptr->tree_type = NONE;
	
	/* process neighbors */
	for(int k = 0; k < 4; k++) {
		x_j = n_i->pos_x + NB::nb_x[k];
		y_j = n_i->pos_y + NB::nb_y[k];
		if( IN_IMAGE(x_j, y_j) ) {
			n_j = getNode(x_j, y_j);
			if( n_j->parent_node == n_i ) {
				n_j->parent_node = ORPHAN;
				orphan_set.push(n_j);
			}
		}
	}
}

void TGraphImage::process_sink_orphan(t_node *n_i)
{
	int x_j, y_j;
	t_node *n_j, *n_tmp;
	bool found_terminal = false;
	int k_min, dist, dist_min = INFINITE_D;
	t_node *nd_min = NULL;
	
	/* try to find a new parent with the same label */
	for(int k = 0; k < 4; k++) {
		x_j = n_i->pos_x + NB::nb_x[k];
		y_j = n_i->pos_y + NB::nb_y[k];
		if( IN_IMAGE(x_j, y_j) ) {
			n_j = getNode(x_j, y_j);
			if( n_j->node_ptr->tree_type == SINK && n_j->t_label == n_i->t_label && n_i->arc_ptr[k]->r_cap > 0 )  {				
				dist = 0;
				n_tmp = n_j;
				while(1) {
					if( n_tmp->parent_node == ORPHAN || n_tmp->parent_node == n_i) {						
						dist = INFINITE_D;
						break;
					}else if(n_tmp->parent_node == TERMINAL) {
						found_terminal = true;						
						break;
					}else {
						n_tmp = n_tmp->parent_node;
						dist++;
					}
				}
				// find the shortest to the path to the terminal
				if(dist == 0) { // to reduce iteration number
					n_i->parent_node = n_j;
					n_i->parc_ind = k;
					return;
				}else if(dist < dist_min) { 
					nd_min = n_j;
					k_min = k;
				}
			}			
		}// end if( IN_IMAGE(x_j, y_j) )
	}// end for(int k = 0; k < 4; k++)

	if( found_terminal ) {
		n_i->parent_node = nd_min;
		n_i->parc_ind = k_min;				
		return;
	}

	/* no new legal parent is found, now n_i is free */
	if(ND_STATE(n_i) != FREE) { // it belonged to an active set or passive set
		/* in act_Sf */
		if( ND_STATE(n_i) == ACTIVE ) 
			if(getTLabel(n_i) == FOREGROUND)
				delete_node(act_Tf, n_i);
			else
				delete_node(act_Tb, n_i);
		else 
			if(getTLabel(n_i) == FOREGROUND)
				delete_node(pas_Tf, n_i);			
			else
				delete_node(pas_Tb, n_i);
	}
	n_i->parent_node = NULL;	
	n_i->node_ptr->tree_type = NONE;

	/* process neighbors */
	for(int k = 0; k < 4; k++) {
		x_j = n_i->pos_x + NB::nb_x[k];
		y_j = n_i->pos_y + NB::nb_y[k];
		if( IN_IMAGE(x_j, y_j) ) {
			n_j = getNode(x_j, y_j);
			if( n_j->parent_node == n_i ) {
				n_j->parent_node = ORPHAN;
				orphan_set.push(n_j);
			}
		}
	}
}

