#include "tgraphimage.h"

TGraphImage::TGraphImage(int width, int height, ErrorFunction errorFun) :
m_width(width), m_height(height)
{
	nodes = new t_node[m_width * m_height];
	for (int i = 0; i < m_width * m_height; i++) {		
		nodes[i].node_ptr = (node*)add_node();
		nodes[i].arc_ptr[0] = nodes[i].arc_ptr[1] = nodes[i].arc_ptr[2] = nodes[i].arc_ptr[3] = NULL;
		nodes[i].pos_x = i / m_height;
		nodes[i].pos_y = i % m_height;		
		nodes[i].t_label = BACKGROUND;	// default: background
	}

	act_Sb = new TNODE_LPQUEUE(- (width + height), (width + height), ACTIVE);
	act_Sf = new TNODE_LPQUEUE(- (width + height), (width + height), ACTIVE);
	act_Tb = new TNODE_LPQUEUE(- (width + height), (width + height), ACTIVE);
	act_Tf = new TNODE_LPQUEUE(- (width + height), (width + height), ACTIVE);

	pas_Sb = new TNODE_LPQUEUE(- (width + height), (width + height), PASSIVE);
	pas_Sf = new TNODE_LPQUEUE(- (width + height), (width + height), PASSIVE);
	pas_Tb = new TNODE_LPQUEUE(- (width + height), (width + height), PASSIVE);
	pas_Tf = new TNODE_LPQUEUE(- (width + height), (width + height), PASSIVE);

	sdmap = new SDistMap(m_width, m_height, errorFun);

	nb_type = NB_FOUR;//NB_EIGHT; //
}

TGraphImage::~TGraphImage(void)
{
	delete[] nodes;

	delete act_Sb;	delete act_Sf;	delete act_Tb;	delete act_Tf;
	delete pas_Sb;	delete pas_Sf;	delete pas_Tb;	delete pas_Tf;

	delete sdmap;
}

/*	
	Add the edge (x1, y1)<->(x2, y2) to the image grid
	it can be only called once for each edge
*/
void TGraphImage::add_edge(int x1, int y1, int x2, int y2, CapType cap)
{
	//Inside the image
	_ASSERT( x1 >= 0 && x1 < m_width && y1 >=0 && y1 < m_height);
	_ASSERT( x2 >= 0 && x2 < m_width && y2 >=0 && y2 < m_height);

    
	arc *a, *a_rev;

	a = arc_block -> New(2);
	a_rev = a + 1;

	a -> sister = a_rev;
	a_rev -> sister = a;
	a -> r_cap = a_rev -> r_cap = cap;
	
	if( x1 == x2 && y1 != y2) {
		if( y1 == y2 - 1) { // 1 in left, 2 in right
			getNode(x1, y1)->arc_ptr[ARC_RIGHT] = a;
			getNode(x2, y2)->arc_ptr[ARC_LEFT]= a_rev; 
		}else if( y1 == y2 + 1) { // 2 in left, 1 in right
			getNode(x1, y1)->arc_ptr[ARC_LEFT] = a;
			getNode(x2, y2)->arc_ptr[ARC_RIGHT] = a_rev;
		}else {
			_ASSERT(false);
		}
	}else if( x1 != x2 && y1 == y2) {
		if( x1 == x2 - 1) { // 1 in up, 2 in down
			getNode(x1, y1)->arc_ptr[ARC_DOWN] = a;
			getNode(x2, y2)->arc_ptr[ARC_UP] = a_rev;
		}else if( x1 == x2 +1) { // 2 in up, 1 in down
			getNode(x1, y1)->arc_ptr[ARC_UP] = a;
			getNode(x2, y2)->arc_ptr[ARC_DOWN] = a_rev;
		}else {
			_ASSERT(false);
		}
	}else{
		_ASSERT(false);
	}
}


/*
	Set the weight between the node (x,y) and two terminals SOURCE and SINK
*/
void TGraphImage::set_tweights(int x, int y, CapType cap_source, CapType cap_sink)
{
	_ASSERT( x >= 0 && x < m_width && y >=0 && y < m_height);

	Graph::set_tweights(getNode(x, y)->node_ptr, cap_source, cap_sink);
}


TGraphImage::TLabel TGraphImage::what_segment(int x, int y)
{
	_ASSERT( x >= 0 && x < m_width && y >=0 && y < m_height);

	return getTLabel(getNode(x, y));
}

Graph::FlowType TGraphImage::maxflow()
{
	maxflow_init();

	return tp_maxflow();
}

/*	
	This can be called multiple times.
    But ensure that those contours do not self-cross or overlap.
	More convenient label setting is to be implemented.
*/
void TGraphImage::add_contour(int *x, int *y, int numOfPnts)
{
	_ASSERT(numOfPnts > 0);

	sdmap->begin_contour(x[0], y[0]);
	for(int i = 0; i < numOfPnts; i++) {
		sdmap->add_point(x[i], y[i]);
	}
	sdmap->end_contour();

	sdmap->code_DMap();
}
